@@ -19,7 +19,7 @@
 
 /* action.h
  *
- * $Id: action.h,v 1.7 2003/11/08 18:24:57 whamann Exp $
+ * $Id: action.h,v 1.12 2007/11/02 10:17:04 krechert Exp $
  *
  * Notice: This header file contains declarations of functions and types that
  * are just used internally. All library functions and types that are supposed
@@ -32,10 +32,13 @@
 #include "ming.h"
 #include "outputblock.h"
 
-SWFAction newSWFAction();
 
-SWFAction newSWFAction_fromOutput(SWFOutput out);
+SWFInitAction newSWFInitAction_MovieClip(SWFMovieClip clip, SWFAction action);
 
 void destroySWFAction(SWFAction action);
+void destroySWFInitAction(SWFInitAction action);
 
+void SWFOutput_writeAction(SWFOutput out, SWFAction action);
+
+SWFMovieClip SWFInitAction_getMovieClip(SWFInitAction action);
 #endif /* SWF_ACTION_H_INCLUDED */
