@@ -17,11 +17,10 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-/* $Id: fromswf.c,v 1.14 2006/02/08 19:09:27 anderson Exp $ */
+/* $Id: fromswf.c,v 1.28 2009/02/16 17:40:41 krechert Exp $ */
 
 
-#include <stdlib.h>
-#include <string.h>
+#include "ming_config.h"
 
 #include "libming.h"
 #include "fromswf.h"
@@ -30,6 +29,15 @@
 #include "output.h"
 #include "error.h"
 
+#ifndef __C2MAN__
+#include <stdlib.h>
+#include <string.h>
+#endif
+
+#if USE_ZLIB
+#include <zlib.h>
+#endif
+
 static void
 writeSWFPrebuiltClipToMethod(SWFBlock block, SWFByteOutputMethod method, void *data)
 {
@@ -47,11 +55,14 @@ completeSWFPrebuiltClip(SWFBlock block)
 }
 
 int
-SWFPrebuiltClipLength(block)
+SWFPrebuiltClipLength(SWFBlock block)
 {	SWFPrebuiltClip clip = (SWFPrebuiltClip)block;
 	return SWFOutput_getLength(clip->display);
 }
 
+/*
+ * destroys a SWFPrebuiltClip instance
+ */
 void
 destroySWFPrebuiltClip(SWFPrebuiltClip clip)
 {	destroySWFOutput(clip->display);
@@ -158,7 +169,7 @@ static void rect(BITS bp)
 	ymin = getbits(bp, nbits);
 	ymax = getbits(bp, nbits);
 	if(verbose)
-		printf("rect %.2f,%.2f %.2f,%.2f\n", xmin/20., ymin/20., xmax/20., ymax/20.);
+		printf("rect %.2f,%.2f %.2f,%.2f\n", xmin/Ming_scale, ymin/Ming_scale, xmax/Ming_scale, ymax/Ming_scale);
 }
 /* matrix */
 static void matrix(BITS bp)
@@ -232,7 +243,7 @@ static void putint4(unsigned char *p, int val)
 struct swfile
 {	char lastch;
 	char bitoff;
-	unsigned char (*readc)();
+	unsigned char (*readc)(struct swfile *);
 
 	char *name;	
 	unsigned char vers[4];
@@ -256,7 +267,6 @@ static void swfseek(struct swfile *sp, int delta)
 {	SWFInput_seek(sp->input, delta, 1);
 }
 
-#include <zlib.h>
 static struct swfile *openswf(SWFInput input)
 {	struct swfile *res = (struct swfile *)malloc(sizeof(struct swfile));
 	SWFInput_read(input, res->vers, 4);
@@ -265,7 +275,9 @@ static struct swfile *openswf(SWFInput input)
 	res->fsize = SWFInput_getUInt32(input);
 	res->compressed = res->vers[0] == 'C';
 	if(res->compressed)
-	{	static z_stream z = {0};
+	{
+#if USE_ZLIB
+		static z_stream z = {0};
 		int len = SWFInput_length(input);
 		unsigned char *zbuf;
 
@@ -278,6 +290,10 @@ static struct swfile *openswf(SWFInput input)
 		inflate(&z, Z_FINISH);
 		inflateEnd(&z);
 		input = newSWFInput_allocedBuffer(zbuf, z.next_out-zbuf);
+#else
+		SWF_error("The SWF to be opened is compressed, but we can't uncompress it (no zlib compiled into this version of Ming).\n");
+		return NULL;
+#endif
 	}
 	res->input = input;
 	// setup to read that rect...
@@ -295,7 +311,7 @@ static struct swfile *openswf(SWFInput input)
 struct swftag
 {	char lastch;
 	char bitoff;
-	unsigned char (*readc)();
+	unsigned char (*readc)(struct swftag *sp);
 
 	short type; int size;
 	unsigned char hdr[6]; short hdrlen;
@@ -358,19 +374,33 @@ static int change_id(TAG tp)
 }
 
 // processing functions for handle()
-static void defineshape();
-static void definetext();
-static void placeobject2();
-static void definebutton();
-static void definebutton2();
-static void definetextfield();
-static void definesprite();
-static void definemorphshape();
-static void exportassets();
-static void definebuttonsound();
-static void fillandlinestyles();
-static void linestyle();
-static void shape();
+static void defineshape(TAG tp, int lev);
+static void definetext(TAG tp, int lev);
+static void placeobject(TAG tp, int lv);
+static void definebutton(TAG tp);
+static void definebutton2(TAG tp);
+static void definetextfield(TAG tp);
+static void definesprite(TAG tp);
+static void definemorphshape(TAG tp,int lev);
+static void exportassets(TAG tp);
+static void definebuttonsound(TAG tp);
+static void fillandlinestyles(TAG tp, int lev);
+static void linestyle(TAG tp, int lev);
+static void shape(TAG tp, int lev);
+static void morphlinestyle2(TAG tp);
+
+static int drop_tag(TAG tp)
+{
+	switch(tp->type)
+	{
+		case SWF_FILEATTRIBUTES:
+		case SWF_METADATA:
+		case SWF_DEFINESCENEANDFRAMEDATA: // only allowed in main timeline -> merge ?
+			return 1;
+		default:
+			return 0;
+	}
+}
 
 static int handle_tag(TAG tp)
 {	int id;
@@ -429,10 +459,14 @@ static int handle_tag(TAG tp)
 				tagnam = "stagProtect"; break;
 			case SWF_PLACEOBJECT2:
 				tagnam = "stagPlaceObject2"; break;
+			case SWF_PLACEOBJECT3:
+				tagnam = "stagPlaceObject3"; break;
 			case SWF_REMOVEOBJECT2:
 				tagnam = "stagRemoveObject2"; break;
 			case SWF_DEFINESHAPE3:
 				tagnam = "stagDefineShape3"; break;
+			case SWF_DEFINESHAPE4:
+				tagnam = "stagDefineShape4"; break;
 			case SWF_DEFINETEXT2:
 				tagnam = "stagDefineText2"; break;
 			case SWF_DEFINEBUTTON2:
@@ -453,8 +487,14 @@ static int handle_tag(TAG tp)
 				tagnam = "stagSoundStreamHead2"; break;
 			case SWF_DEFINEMORPHSHAPE:
 				tagnam = "stagDefineMorphShape"; break;
+			case SWF_DEFINEMORPHSHAPE2:
+				tagnam = "stagDefineMorphShape2"; break;
 			case SWF_DEFINEFONT2:
 				tagnam = "stagDefineFont2"; break;
+			case SWF_DEFINEFONT3:
+				tagnam = "stagDefineFont3"; break;
+			case SWF_DEFINEFONTALIGNZONES:
+				tagnam = "stagDefineFontAlignZones"; break;
 			case SWF_EXPORTASSETS:
 				tagnam = "stagExportAssets"; break;
 			/*case stagClipInit:	
@@ -465,6 +505,8 @@ static int handle_tag(TAG tp)
 				tagnam = "stagDefineVideoStrem"; break;
 			case SWF_VIDEOFRAME:
 				tagnam = "stagVideoFrame"; break;
+			case SWF_INITACTION:
+				tagnam = "stagInitAction"; break;
 		}
 		printf("tag %d %s\n", tp->type, tagnam);
 	}
@@ -481,6 +523,9 @@ static int handle_tag(TAG tp)
 		case SWF_SOUNDSTREAMHEAD2:
 		case SWF_SOUNDSTREAMBLOCK:
 		case SWF_JPEGTABLES:
+		case SWF_SYMBOLCLASS:
+		case SWF_DOABC:
+		case SWF_DEFINEFONTNAME:
 		/*case stagClipInit:*/
 			break;
 		/* simple tags - one id at beginning */
@@ -498,10 +543,13 @@ static int handle_tag(TAG tp)
 		case SWF_STARTSOUND:
 		case SWF_NAMECHARACTER:
 		case SWF_DEFINEFONT2:
+		case SWF_DEFINEFONT3:
+		case SWF_DEFINEFONTALIGNZONES:
 		case SWF_DEFINEBUTTONCXFORM:
 		case SWF_DEFINEVIDEOSTREAM:
 		case SWF_VIDEOFRAME:
 		case SWF_INITACTION:
+		case SWF_DEFINEBINARYDATA:
 		/*case SWF_NEWFONTINFO:*/
 			id = change_id(tp);
 			if(verbose)
@@ -518,11 +566,17 @@ static int handle_tag(TAG tp)
 			defineshape(tp, 2);
 			break;
 		case SWF_PLACEOBJECT2:
-			placeobject2(tp);
+			placeobject(tp, 2);
+			break;
+		case SWF_PLACEOBJECT3:
+			placeobject(tp, 3);
 			break;
 		case SWF_DEFINESHAPE3:
 			defineshape(tp, 3);
 			break;
+		case SWF_DEFINESHAPE4:
+			defineshape(tp, 4);
+			break;
 		case SWF_DEFINETEXT2:
 			definetext(tp, 2);
 			break;
@@ -539,7 +593,10 @@ static int handle_tag(TAG tp)
 			definesprite(tp);
 			break;
 		case SWF_DEFINEMORPHSHAPE:
-			definemorphshape(tp);
+			definemorphshape(tp, 1);
+			break;
+		case SWF_DEFINEMORPHSHAPE2:
+			definemorphshape(tp, 2);
 			break;
 		case SWF_EXPORTASSETS:
 			exportassets(tp);
@@ -561,12 +618,14 @@ static int handle_tag(TAG tp)
 		case SWF_SOUNDSTREAMBLOCK:
 		case SWF_DEFINEVIDEOSTREAM:
 		case SWF_VIDEOFRAME:
+		case SWF_INITACTION:
 		case SWF_FREECHARACTER:
 		case SWF_PLACEOBJECT:
 		case SWF_REMOVEOBJECT:
 		case SWF_STARTSOUND:
 		case SWF_NAMECHARACTER:
 		case SWF_PLACEOBJECT2:
+		case SWF_PLACEOBJECT3:
 			displaylist = 1;
 	}
 	return displaylist;
@@ -670,7 +729,7 @@ static void shaperecord(TAG tp, int nfillbits, int nlinebits, int lev)
 	}
 }
 
-static void gradient(TAG tp, int alpha)
+static void gradient(TAG tp, int alpha, int cod)
 {	int n, ngrad, r1;
 	
 	alignbits(tp);
@@ -681,6 +740,8 @@ static void gradient(TAG tp, int alpha)
 			printf("ratio %d\n", r1);
 		if(alpha) rgba((BITS) tp); else rgb((BITS) tp);
 	}
+	if (cod == 0x13)
+		readint2((BITS) tp);
 }
 
 static void fillstyle(TAG tp, int lev)
@@ -696,25 +757,31 @@ static void fillstyle(TAG tp, int lev)
 			if(verbose) printf("linear gradient fill\n"); break;
 		case 0x12:
 			if(verbose) printf("radial gradient fill:\n"); break;
+		case 0x13:
+			if(verbose) printf("focal gradient fill:\n"); break;
 		case 0x40:
 			if(verbose) printf("tiled bitmap fill:\n"); break;
 		case 0x41:
 			if(verbose) printf("clipped bitmap fill\n"); break;
+		case 0x42:
+			if(verbose) printf("tilled bitmap fill with hard edges\n"); break;
+		case 0x43:
+			if(verbose) printf("clipped bitmap fill with hard edges\n"); break;
 	}
 	if(cod == 0)
-		if(lev == 3) rgba((BITS) tp); else rgb((BITS) tp);
-	else if(cod == 0x10 || cod == 0x12)
+		if(lev >= 3) rgba((BITS) tp); else rgb((BITS) tp);
+	else if(cod == 0x10 || cod == 0x12 || cod == 0x13)
 	{	matrix((BITS) tp);
-		gradient(tp, lev == 3);
+		gradient(tp, lev >= 3, cod);
 	}
-	else if(cod == 0x40 || cod == 0x41)
+	else if(cod == 0x40 || cod == 0x41 || cod == 0x42 || cod == 0x43)
 	{	id = change_id(tp);
 		if(verbose)
 			printf("fill with id %d\n", id);
 		matrix((BITS) tp);
 	}
 	else
-		printf("UNEXPEDCED %x\n", cod);
+		printf("%s:%d: UNEXPEDCED %x\n", __FILE__, __LINE__, cod);
 }
 
 static void morphgradient(TAG tp)
@@ -747,22 +814,28 @@ static void morphfillstyle(TAG tp)
 			if(verbose) printf("linear gradient fill\n"); break;
 		case 0x12:
 			if(verbose) printf("radial gradient fill:\n"); break;
+		case 0x13:
+			if(verbose) printf("focal gradient fill:\n"); break;
 		case 0x40:
 			if(verbose) printf("tiled bitmap fill:\n"); break;
 		case 0x41:
 			if(verbose) printf("clipped bitmap fill\n"); break;
+		case 0x42:
+			if(verbose) printf("tilled bitmap fill with hard edges\n"); break;
+		case 0x43:
+			if(verbose) printf("clipped bitmap fill with hard edges\n"); break;
 	}
 	if(cod == 0)
 	{	rgba((BITS) tp);
 		rgba((BITS) tp);
 	}
-	else if(cod == 0x10 || cod == 0x12)
+	else if(cod == 0x10 || cod == 0x12 || cod == 0x13)
 	{	matrix((BITS) tp);
 		alignbits(tp);
 		matrix((BITS) tp);
 		morphgradient(tp);
 	}
-	else if(cod == 0x40 || cod == 0x41)
+	else if(cod == 0x40 || cod == 0x41 || cod == 0x42 || cod == 0x43)
 	{	id = change_id(tp);
 		if(verbose) printf("fill with id %d\n", id);
 		matrix((BITS) tp);
@@ -770,7 +843,7 @@ static void morphfillstyle(TAG tp)
 		matrix((BITS) tp);
 	}
 	else
-		printf("UNEXPEDCED %x\n", cod);
+		printf("%s:%d: UNEXPEDCED %x\n", __FILE__, __LINE__, cod);
 }
 
 static void fillandlinestyles(TAG tp, int lev)
@@ -791,7 +864,10 @@ static void fillandlinestyles(TAG tp, int lev)
 	if(verbose)
 		printf ("%d line styles\n", nline);
 	for(n = 0 ; n < nline ; n++)
-		linestyle(tp, lev);
+		if (lev == 4)
+			morphlinestyle2(tp);
+		else
+			linestyle(tp, lev);
 }
 
 static void linestyle(TAG tp, int lev)
@@ -799,7 +875,7 @@ static void linestyle(TAG tp, int lev)
 	w = readint2((BITS) tp);
 	if(verbose)
 		printf ("w %d\n", w);
-	if(lev == 3) rgba((BITS) tp); else rgb((BITS) tp);
+	if(lev >= 3) rgba((BITS) tp); else rgb((BITS) tp);
 }
 
 static void morphlinestyle(TAG tp)
@@ -811,13 +887,55 @@ static void morphlinestyle(TAG tp)
 	rgba((BITS) tp); rgba((BITS) tp);
 }
 
+static void morphlinestyle2(TAG tp)
+{	unsigned short w1;
+	int join_style;
+	int has_fill;
+	int is_morph = (tp->type == SWF_DEFINEMORPHSHAPE2);
+	w1 = readint2((BITS) tp);
+	if (is_morph) {
+		unsigned short w2;
+		w2 = readint2((BITS) tp);
+		if(verbose)
+			printf("w %d => %d\n", w1, w2);
+	} else {
+		if(verbose)
+			printf("w %d\n", w1);
+	}
+	getbits((BITS) tp, 2);
+	join_style = getbits((BITS) tp, 2);
+	has_fill = getbits((BITS) tp, 1);
+	getbits((BITS) tp, 11);
+	if (join_style == 2) {
+		readint2((BITS) tp);
+	} else {
+		if (has_fill) {
+			if (is_morph)
+				morphfillstyle(tp);
+			else
+				fillstyle(tp, 4);
+		} else {
+			rgba((BITS) tp);
+			if (is_morph)
+				rgba((BITS) tp);
+		}
+	}
+}
+
 static void defineshape(TAG tp, int lev)
 {	unsigned short id;
 
 	id = change_id(tp);
 	if(verbose)
 		printf("shape %d\n", id);
+	alignbits(tp);
 	rect((BITS) tp);
+	if(lev == 4) {
+		alignbits(tp);
+		rect((BITS) tp);
+		alignbits(tp);
+		tp->readc(tp);
+	}
 	fillandlinestyles(tp, lev);
 	shape(tp, lev);
 }
@@ -830,7 +948,7 @@ static void shape(TAG tp, int lev)
 	shaperecord(tp, nfillbits, nlinebits,lev);
 }
 
-static void definemorphshape(TAG tp)
+static void definemorphshape(TAG tp, int lev)
 {	unsigned short id, fcnt, lcnt;
 	unsigned long loff;
 	unsigned char *endp;
@@ -842,6 +960,11 @@ static void definemorphshape(TAG tp)
 	rect((BITS) tp);
 	alignbits(tp);
 	rect((BITS) tp);
+	if (lev == 2) {
+		rect((BITS) tp);
+		rect((BITS) tp);
+		tp->readc(tp);
+	}
 	loff = readint4((BITS) tp);
 	endp = tp->datptr + loff;
 	fcnt = tp->readc(tp);
@@ -860,7 +983,10 @@ static void definemorphshape(TAG tp)
 		printf ("%d line styles\n", lcnt);
 	for(n = 0 ; n < lcnt ; n++)
 	{	alignbits(tp);
-		morphlinestyle(tp);
+		if (lev == 2)
+			morphlinestyle2(tp);
+		else
+			morphlinestyle(tp);
 	}
 /*	for(n = 0 ; tp->datptr < endp ; n++)
 	{	alignbits(tp);
@@ -930,9 +1056,16 @@ static void definetext(TAG tp, int lev)
 	}
 }
 
-static void placeobject2(TAG tp)
+static void placeobject(TAG tp, int lv)
 {	int hasname, hasratio, hascxform, hasmatrix, haschar, hasmove, hasactions, hasmask;
 	short depth, charid;
+	int hasfilters, hasbitmapcaching, hasblendmode;
+	if (lv == 3) {
+		getbits((BITS)tp, 5);
+		hasbitmapcaching = getbits((BITS)tp, 1);
+		hasblendmode = getbits((BITS)tp, 1);
+		hasfilters = getbits((BITS)tp, 1);
+	}
 	hasactions = getbits((BITS)tp, 1);
 	hasmask = getbits((BITS)tp, 1);
 	hasname = getbits((BITS)tp, 1);
@@ -1138,6 +1271,9 @@ static void definebuttonsound(TAG tp)
 
 extern int SWF_gNumCharacters;
 
+/* 
+ * load and create a SWF File as MovieClip 
+ */
 SWFPrebuiltClip
 newSWFPrebuiltClip_fromInput(SWFInput input)
 {	SWFPrebuiltClip clip;
@@ -1145,9 +1281,10 @@ newSWFPrebuiltClip_fromInput(SWFInput input)
 	SWFPrebuilt deps;
 	TAG tp;
 	struct swfile *swf;
-	int type, todisplay;
+	int type = 0, todisplay;
 	
 	swf = openswf(input);
+	if ( ! swf ) return NULL;
 	clip = newSWFPrebuiltClip();
 	clip->frames = swf->frames;
 	display = clip->display;
@@ -1155,6 +1292,7 @@ newSWFPrebuiltClip_fromInput(SWFInput input)
 	SWFCharacter_addDependency((SWFCharacter)clip, (SWFCharacter)deps);
 	defines = deps->defines;
 	idoffset = SWF_gNumCharacters;
+	maxid = SWF_gNumCharacters;
 	// read the swf file here
 	tp = readtag_file(swf);
 		if(tp->type != SWF_SETBACKGROUNDCOLOR)
@@ -1164,8 +1302,15 @@ newSWFPrebuiltClip_fromInput(SWFInput input)
 	free(tp);
 	do
 	{	tp = readtag_file(swf);
-		todisplay = handle_tag(tp);
 		type = tp->type;
+		if(drop_tag(tp))
+		{
+			if(tp->alloced)
+				free(tp->datbuf);
+			free(tp);
+			continue;
+		}
+		todisplay = handle_tag(tp);
 		out = todisplay ? display : defines;
 		SWFOutput_writeBuffer(out, tp->hdr, tp->hdrlen);
 		if(tp->size)
@@ -1175,12 +1320,15 @@ newSWFPrebuiltClip_fromInput(SWFInput input)
 		free(tp);
 	} while(type);
 	if(swf->compressed)
-		destroySWFInput(swf->input);	
+		destroySWFInput(swf->input);
 	SWF_gNumCharacters = maxid + 1;
 	CHARACTERID(clip) = SWF_gNumCharacters++;
 	return clip;
 }
 
+/* 
+ * load SWF file which can be used as a MovieClip
+ */
 SWFPrebuiltClip
 newSWFPrebuiltClip_fromFile(const char *filename)
 {	FILE *fp;
