@@ -17,11 +17,13 @@
 		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA
 */
 
-/* $Id: text.c,v 1.29 2004/06/27 07:18:56 whamann Exp $ */
+/* $Id: text.c,v 1.43 2009/12/01 20:09:44 strk Exp $ */
 
+#ifndef __C2MAN__
 #include <stdlib.h>
 #include <string.h>
 #include <math.h>
+#endif
 
 #include "text.h"
 #include "rect.h"
@@ -53,15 +55,12 @@ struct SWFTextRecord_s
 	struct SWFTextRecord_s *next;
 
 	byte flags;
-	BOOL isBrowserFont;
-
 	// if it's not a browser font, is it a font or a fontchar?
 	BOOL isResolved;
 
 	union
 	{
 		SWFFont font;
-		SWFBrowserFont browserFont;
 		SWFFontCharacter fontchar;
 	} font;
 
@@ -83,20 +82,20 @@ struct SWFTextRecord_s
 	unsigned short* string;
 
 	int* advance;
+	int advAllocated;
 	int nAdvanceBits;
 };
 
 
 static void
-writeSWFTextToMethod(SWFBlock block,
-										 SWFByteOutputMethod method, void *data)
+writeSWFTextToMethod(SWFBlock block, SWFByteOutputMethod method, void *data)
 {
 	SWFText text = (SWFText)block;
 	int length = 0;
 	SWFOutput out;
 
 	if ( text->matrix == NULL )
-		text->matrix = newSWFMatrix(0, 0, 0, 0, 0, 0);
+		text->matrix = newSWFMatrix(1.0, 0, 0, 1.0, 0, 0);
 
 	length += (SWFMatrix_numBits(text->matrix)+7)/8;
 	length += (SWFRect_numBits(CHARACTER(text)->bounds)+7)/8;
@@ -166,8 +165,14 @@ destroySWFText(SWFText text)
 SWFText
 newSWFText()
 {
+	SWFRect temp_rect;
+
 	SWFText text = (SWFText)malloc(sizeof(struct SWFText_s));
 
+	/* If malloc failed, return NULL to signify this */
+	if (NULL == text)
+		return NULL;
+
 	SWFCharacterInit((SWFCharacter)text);
 
 	CHARACTERID(text) = ++SWF_gNumCharacters;
@@ -177,9 +182,27 @@ newSWFText()
 	BLOCK(text)->complete = completeSWFText;
 	BLOCK(text)->dtor = (destroySWFBlockMethod) destroySWFText;
 
-	CHARACTER(text)->bounds = newSWFRect(0,0,0,0);
+	temp_rect = newSWFRect(0,0,0,0);
+
+	/* If newSWFRect() failed, return NULL to signify this */
+	if (NULL == temp_rect)
+	{
+		free(text);
+		return NULL;
+	}
+
+	CHARACTER(text)->bounds = temp_rect;
 
 	text->out = newSWFOutput();
+
+	/* If newSWFOutput() failed, return NULL to signify this */
+	if (NULL == text->out)
+	{
+		destroySWFRect(temp_rect);
+		free(text);
+		return NULL;
+	}
+
 	text->currentRecord = NULL;
 	text->initialRecord = NULL;
 	text->matrix = NULL;
@@ -198,6 +221,11 @@ SWFText
 newSWFText2()
 {
 	SWFText text = newSWFText();
+
+	/* If malloc failed, return NULL to signify this */
+	if (NULL == text)
+		return NULL;
+
 	BLOCK(text)->type = SWF_DEFINETEXT2;
 	return text;
 }
@@ -207,7 +235,13 @@ SWFTextRecord
 SWFText_addTextRecord(SWFText text)
 {
 	SWFTextRecord textRecord = (SWFTextRecord)malloc(sizeof(struct SWFTextRecord_s));
-	SWFTextRecord current = text->currentRecord;
+	SWFTextRecord current = NULL;
+
+	/* If malloc failed, return NULL to signify this */
+	if (NULL == text)
+		return NULL;
+
+	current = text->currentRecord;
 
 	textRecord->flags = 0;
 	textRecord->string = NULL;
@@ -216,11 +250,11 @@ SWFText_addTextRecord(SWFText text)
 	textRecord->x = 0;
 	textRecord->y = 0;
 	textRecord->advance = NULL;
+	textRecord->advAllocated = 0;
 	textRecord->nAdvanceBits = 0;
 
 	if ( current == NULL )
 	{
-		textRecord->isBrowserFont = FALSE;
 		textRecord->isResolved = FALSE;
 		textRecord->font.font = NULL;
 		textRecord->spacing = 0;
@@ -234,7 +268,6 @@ SWFText_addTextRecord(SWFText text)
 	}
 	else
 	{
-		textRecord->isBrowserFont = current->isBrowserFont;
 		textRecord->isResolved = current->isResolved;
 		textRecord->font = current->font;
 		textRecord->spacing = current->spacing;
@@ -259,10 +292,11 @@ destroySWFTextRecord(SWFTextRecord record)
 	if ( record->string != NULL )
 		free(record->string);
 
-	if ( record->advance != NULL )
+	if ( record->advance != NULL && record->advAllocated)
 		free(record->advance);
 
-	free(record);
+	if ( record != NULL )
+		free(record);
 }
 
 
@@ -291,7 +325,7 @@ SWFTextRecord_getNextRecord(SWFTextRecord record)
 SWFFont
 SWFTextRecord_getUnresolvedFont(SWFTextRecord record)
 {
-	if ( !record->isBrowserFont && !record->isResolved )
+	if ( !record->isResolved )
 		return record->font.font;
 	
 	return NULL;
@@ -311,12 +345,21 @@ int
 SWFText_getScaledStringWidth(SWFText text, const char *string)
 {
 	SWFFont font;
-	int height = text->currentRecord->height;
+	int height;
 	unsigned short* widestr;
 	int len = strlen(string);
 	int n, ret;
+	
+	if(text->currentRecord == NULL)
+		return -1;
 
+	height = text->currentRecord->height;
 	widestr = (unsigned short *)malloc(2 * len);
+
+	/* If malloc failed, return -1 to signify this */
+	if (NULL == text)
+		return -1;
+
 	for(n = 0 ; n < len ; n++)
 		widestr[n] = (unsigned char)string[n];
 
@@ -325,10 +368,7 @@ SWFText_getScaledStringWidth(SWFText text, const char *string)
 	else
 		font = text->currentRecord->font.font;
 
-	if ( text->currentRecord->isBrowserFont )
-		ret = 0;
-	else
-		ret = SWFFont_getScaledWideStringWidth(font, widestr, len) * height / 1024;
+	ret = SWFFont_getScaledWideStringWidth(font, widestr, len) * height / 1024;
 	free(widestr);
 	return ret;
 }
@@ -337,20 +377,21 @@ int
 SWFText_getScaledUTF8StringWidth(SWFText text, const char *string)
 {
 	SWFFont font;
-	int height = text->currentRecord->height;
+	int height;
 	unsigned short* widestr;
 	int len, ret;
 
+	if(text->currentRecord == NULL)
+		return -1;
+
+	height = text->currentRecord->height;
 	len = UTF8ExpandString(string, &widestr);
 	if ( text->currentRecord->isResolved )
 		font = SWFFontCharacter_getFont(text->currentRecord->font.fontchar);
 	else
 		font = text->currentRecord->font.font;
 
-	if ( text->currentRecord->isBrowserFont )
-		ret = 0;
-	else
-		ret = SWFFont_getScaledWideStringWidth(font, widestr, len) * height / 1024;
+	ret = SWFFont_getScaledWideStringWidth(font, widestr, len) * height / 1024;
 	free(widestr);
 	return ret;
 }
@@ -359,46 +400,52 @@ int
 SWFText_getScaledWideStringWidth(SWFText text, const unsigned short *string)
 {
 	SWFFont font;
-	int height = text->currentRecord->height;
+	int height;
 	int len;
 
-	for(len = 0 ; *string ; len++)
+	if(text->currentRecord == NULL)
+		return -1;
+	
+	height = text->currentRecord->height;
+
+	for(len = 0 ; *(string + len); len++)
 		;
 	if ( text->currentRecord->isResolved )
 		font = SWFFontCharacter_getFont(text->currentRecord->font.fontchar);
 	else
 		font = text->currentRecord->font.font;
 
-	if ( text->currentRecord->isBrowserFont )
-		return 0;
-	else
-		return SWFFont_getScaledWideStringWidth(font, string, len) * height / 1024;
+	return SWFFont_getScaledWideStringWidth(font, string, len) * height / 1024;
 }
 
 
 short
 SWFText_getScaledAscent(SWFText text)
 {
-	SWFFont font = text->currentRecord->font.font;
-	int height = text->currentRecord->height;
+	SWFFont font;
+	int height;
 
-	if ( text->currentRecord->isBrowserFont )
-		return 0;
-	else
-		return SWFFont_getScaledAscent(font) * height / 1024;
+	if(text->currentRecord == NULL)
+		return -1;
+
+	font = text->currentRecord->font.font;
+	height = text->currentRecord->height;
+	return SWFFont_getScaledAscent(font) * height / 1024;
 }
 
 
 short
 SWFText_getScaledDescent(SWFText text)
 {
-	SWFFont font = text->currentRecord->font.font;
-	int height = text->currentRecord->height;
+	SWFFont font;
+	int height;
 
-	if ( text->currentRecord->isBrowserFont )
-		return 0;
-	else
-		return SWFFont_getScaledDescent(font) * height / 1024;
+	if(text->currentRecord == NULL)
+		return -1;
+
+	font = text->currentRecord->font.font;
+	height = text->currentRecord->height;
+	return SWFFont_getScaledDescent(font) * height / 1024;
 }
 
 
@@ -408,31 +455,22 @@ SWFText_getScaledLeading(SWFText text)
 	SWFFont font = text->currentRecord->font.font;
 	int height = text->currentRecord->height;
 
-	if ( text->currentRecord->isBrowserFont )
-		return 0;
-	else
-		return SWFFont_getScaledLeading(font) * height / 1024;
+	return SWFFont_getScaledLeading(font) * height / 1024;
 }
 
-
 void 
-SWFText_setFont(SWFText text, void* font)
+SWFText_setFont(SWFText text, SWFFont font)
 {
 	SWFTextRecord textRecord = text->currentRecord;
-
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
 	textRecord->flags |= SWF_TEXT_HAS_FONT;
-
-	/* XXX */
-	textRecord->isBrowserFont = (BLOCK(font)->type == SWF_DEFINEEDITTEXT);
-
-	if ( textRecord->isBrowserFont )
-		SWF_error("cannot use browser font for SWFText");
-//		textRecord->font.browserFont = (SWFBrowserFont)font;
-	else
-		textRecord->font.font = (SWFFont)font;
+	textRecord->font.font = font;
 }
 
 
@@ -460,6 +498,10 @@ SWFText_setScaledSpacing(SWFText text, int spacing)
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
 	textRecord->spacing = spacing;
 }
 
@@ -472,6 +514,10 @@ SWFText_setColor(SWFText text, byte r, byte g, byte b, byte a)
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
 	textRecord->flags |= SWF_TEXT_HAS_COLOR;
 	textRecord->r = r;
 	textRecord->g = g;
@@ -488,6 +534,10 @@ SWFText_scaledMoveTo(SWFText text, int x, int y)
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
 	if ( x != 0 )
 	{
 		textRecord->flags |= SWF_TEXT_HAS_X;
@@ -511,10 +561,6 @@ SWFText_addWideString(SWFText text, const unsigned short* widestring,
 											int len, int* advance)
 {
 	SWFTextRecord textRecord = text->currentRecord;
-	SWFFont font = textRecord->font.font;
-
-	if ( font == NULL )
-		SWF_error("font must be set before calling addString");
 
 	/* marginally sloppy to tack on a new record,
 		 but I don't want to deal with concats */
@@ -522,9 +568,24 @@ SWFText_addWideString(SWFText text, const unsigned short* widestring,
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
+	if ( textRecord->font.font == NULL )
+		SWF_error("font must be set before calling addString");
+
 	textRecord->advance = advance;
 	textRecord->strlen = len;
 	textRecord->string = (unsigned short*)malloc(sizeof(unsigned short) * len);
+
+	/* If malloc failed, return early */
+	if (NULL == textRecord->string)
+	{
+		destroySWFTextRecord(textRecord);
+		return;
+	}
+
 	memcpy(textRecord->string, widestring, sizeof(unsigned short) * len);
 }
 
@@ -536,10 +597,6 @@ SWFText_addUTF8String(SWFText text, const char* string, int* advance)
 	int len = UTF8ExpandString(string, &widestring);
 
 	SWFTextRecord textRecord = text->currentRecord;
-	SWFFont font = textRecord->font.font;
-
-	if ( font == NULL )
-		SWF_error("font must be set before calling addString");
 
 	/* marginally sloppy to tack on a new record,
 		 but I don't want to deal with concats */
@@ -547,6 +604,13 @@ SWFText_addUTF8String(SWFText text, const char* string, int* advance)
 	if ( textRecord == NULL || textRecord->string != NULL )
 		textRecord = SWFText_addTextRecord(text);
 
+	/* If SWFText_addTextRecord() failed, return early */
+	if (NULL == textRecord)
+		return;
+
+	if ( textRecord->font.font == NULL )
+		SWF_error("font must be set before calling addString");
+
 	textRecord->advance = advance;
 	textRecord->strlen = len;
 	textRecord->string = widestring;
@@ -559,6 +623,11 @@ SWFText_addString(SWFText text, const char* string, int* advance)
 	int len = strlen(string);
 	int i;
 	unsigned short* widestring = (unsigned short*) malloc(len * sizeof(unsigned short) );
+
+	/* If malloc() failed, return early */
+	if (NULL == widestring)
+		return;
+
 	for ( i = 0; i < len; ++i )
 		widestring[i] = string[i] & 0xff;
 	SWFText_addWideString(text, widestring, len, advance);
@@ -580,14 +649,18 @@ SWFTextRecord_computeAdvances(SWFTextRecord textRecord)
 	if(!len) return;	// do not try to calculate 1st char of null string
 
 	/* compute advances (spacing) from spacing, advance and kern table */
-
 	if ( textRecord->advance == NULL )
 	{
 		textRecord->advance = (int*)malloc(sizeof(int) * len);
+		/* If malloc() failed, return early */
+		if (NULL == textRecord->advance)
+			return;
+
+		textRecord->advAllocated = 1;
 		memset(textRecord->advance, 0, sizeof(int) * len);
 	}
 
-	glyph = SWFFontCharacter_getGlyphCode(fontchar, widestring[0]);
+	glyph = SWFFont_findGlyphCode(font, widestring[0]);
 
 	for ( i=0; i<len; ++i )
 	{
@@ -600,7 +673,7 @@ SWFTextRecord_computeAdvances(SWFTextRecord textRecord)
 		/* get kerning from font's kern table */
 
 		if ( i < len-1 )
-		{	nextglyph = SWFFontCharacter_getGlyphCode(fontchar, widestring[i+1]);
+		{	nextglyph = SWFFont_findGlyphCode(font, widestring[i+1]);
 			adv += SWFFont_getCharacterKern(font, glyph, nextglyph);
 			glyph = nextglyph;
 		}
@@ -611,7 +684,7 @@ SWFTextRecord_computeAdvances(SWFTextRecord textRecord)
 		textRecord->advance[i] = adv * textRecord->height / 1024;
 
 		textRecord->nAdvanceBits = max(textRecord->nAdvanceBits,
-																	 SWFOutput_numSBits(textRecord->advance[i]));
+			SWFOutput_numSBits(textRecord->advance[i]));
 
 	}
 }
@@ -637,17 +710,9 @@ SWFText_resolveCodes(SWFText text)
 		text->nAdvanceBits = max(text->nAdvanceBits, textRecord->nAdvanceBits);
 
 		if ( textRecord->flags & SWF_TEXT_HAS_FONT )
-		{
-			if ( textRecord->isBrowserFont )
-			{
-				/* XXX - assume browser fonts have 8bit glyph table? */
-				nGlyphBits = max(nGlyphBits, 8);
-			}
-			else
-			{
-				int fontGlyphs = SWFFontCharacter_getNGlyphs(textRecord->font.fontchar);
-				nGlyphBits = max(nGlyphBits, SWFOutput_numBits(fontGlyphs-1));
-			}
+		{	
+			int fontGlyphs = SWFFontCharacter_getNGlyphs(textRecord->font.fontchar);
+			nGlyphBits = max(nGlyphBits, SWFOutput_numBits(fontGlyphs-1));
 		}
 
 		textRecord = textRecord->next;
@@ -657,6 +722,8 @@ SWFText_resolveCodes(SWFText text)
 
 	while ( textRecord != NULL )
 	{
+		SWFFontCharacter fontchar = NULL;
+		SWFFont font = NULL;
 		oldRecord = textRecord;
 
 		if ( textRecord->string == NULL || textRecord->strlen == 0 )
@@ -674,12 +741,7 @@ SWFText_resolveCodes(SWFText text)
 		SWFOutput_writeUInt8(out, textRecord->flags | SWF_TEXT_STATE_CHANGE);
 
 		if ( textRecord->flags & SWF_TEXT_HAS_FONT )
-		{
-			if ( textRecord->isBrowserFont )
-				SWFOutput_writeUInt16(out, CHARACTERID(textRecord->font.browserFont));
-			else
-				SWFOutput_writeUInt16(out, CHARACTERID(textRecord->font.fontchar));
-		}
+			SWFOutput_writeUInt16(out, CHARACTERID(textRecord->font.fontchar));
 
 		if ( textRecord->flags & SWF_TEXT_HAS_COLOR )
 		{
@@ -718,83 +780,50 @@ SWFText_resolveCodes(SWFText text)
 
 		SWFOutput_writeUInt8(out, len);
 
-		/* XXX - er, browser fonts in text objects crash the player..
-			 Maybe because there's no definefontinfo block? */
+		fontchar = textRecord->font.fontchar;
+		font = SWFFontCharacter_getFont(fontchar);
 
-		if ( textRecord->isBrowserFont )
+		if ( font == NULL )
+			SWF_error("Couldn't find font");
+
+		for ( i=0; i<len; ++i )
 		{
-			for ( i=0; i<len; ++i )
-			{
-				SWFOutput_writeBits(out, textRecord->string[i], nGlyphBits);
-				SWFOutput_writeBits(out, textRecord->advance[i], text->nAdvanceBits);
+			SWFRect glyphBounds;
+			int minX, maxX, minY, maxY;
 
-				/* XXX - fudging the text character bounds since we don't
-					 have font metrics */
+			unsigned short font_glyphcode =
+				SWFFont_findGlyphCode(font, textRecord->string[i]);
+			glyphBounds = SWFFont_getGlyphBounds(font,font_glyphcode);
+			SWFRect_getBounds(glyphBounds, &minX, &maxX, &minY, &maxY);
 
-				if ( CHARACTER(text)->bounds )
-				{
-					SWFRect_includePoint(CHARACTER(text)->bounds, curX, curY, 0);
-					SWFRect_includePoint(CHARACTER(text)->bounds,
-															 curX + curH, curY + curH, 0);
-				}
-				else
-				{
-					CHARACTER(text)->bounds =
-						newSWFRect(curX, curX + curH, curY, curY + curH);
-				}
-
-				curX += curH;
+			int fontchar_glyphcode =
+				SWFFontCharacter_findGlyphCode(fontchar, textRecord->string[i]);
+			if (fontchar_glyphcode < 0) {
+				SWF_error("SWFText_resolveCodes: no suitable glyph available (in dumped font)");
 			}
-		}
-		else
-		{
-			SWFFontCharacter fontchar = textRecord->font.fontchar;
-			SWFFont font = SWFFontCharacter_getFont(fontchar);
+			SWFOutput_writeBits(out, (unsigned short) fontchar_glyphcode,	nGlyphBits);
+			SWFOutput_writeBits(out, textRecord->advance[i], text->nAdvanceBits);
 
-			if ( font == NULL )
-				SWF_error("Couldn't find font");
-
-			for ( i=0; i<len; ++i )
+			if ( CHARACTER(text)->bounds )
 			{
-				SWFRect glyphBounds;
-				int minX, maxX, minY, maxY;
-
-				unsigned short code =
-					SWFFontCharacter_getGlyphCode(fontchar, textRecord->string[i]);
-
-				glyphBounds = SWFFont_getGlyphBounds(font, code);
-
-				SWFRect_getBounds(glyphBounds, &minX, &maxX, &minY, &maxY);
-
-				SWFOutput_writeBits(out, textRecord->string[i],	nGlyphBits);
-				SWFOutput_writeBits(out, textRecord->advance[i], text->nAdvanceBits);
-
-				if ( CHARACTER(text)->bounds )
-				{
-					SWFRect_includePoint(CHARACTER(text)->bounds,
-															 curX + minX * curH / 1024,
-															 curY + minY * curH / 1024, 0);
-
-					SWFRect_includePoint(CHARACTER(text)->bounds,
-															 curX + maxX * curH / 1024,
-															 curY + maxY * curH / 1024, 0);
-				}
-				else
-				{
-					CHARACTER(text)->bounds =
-						newSWFRect(curX + minX * curH /1024, curX + maxX * curH /1024,
-											 curY + minY * curH /1024, curY + maxY * curH /1024);
-				}
-
-				if ( textRecord->advance != NULL )
-					curX += textRecord->advance[i];
+				SWFRect_includePoint(CHARACTER(text)->bounds, curX + minX * curH / 1024, 
+				                     curY + minY * curH / 1024, 0);
+				SWFRect_includePoint(CHARACTER(text)->bounds, curX + maxX * curH / 1024, 
+				                     curY + maxY * curH / 1024, 0);
+			}
+			else
+			{
+				CHARACTER(text)->bounds = newSWFRect(curX + minX * curH /1024, 
+					curX + maxX * curH /1024, curY + minY * curH /1024, 
+					curY + maxY * curH /1024);
 			}
-		}
 
+			if ( textRecord->advance != NULL )
+				curX += textRecord->advance[i];
+		}
 		textRecord = textRecord->next;
 		destroySWFTextRecord(oldRecord);
 	}
-
 	SWFOutput_writeUInt8(out, 0); /* end text records */
 
 	text->nGlyphBits = nGlyphBits;
