@@ -19,7 +19,7 @@
 
 /* font_util.h
  * 
- * $Id: font_util.h,v 1.5 2003/12/14 10:07:14 whamann Exp $
+ * $Id: font_util.h,v 1.6 2008/01/04 20:59:15 krechert Exp $
  * 
  * Notice: This header file contains declarations of functions and types that
  * are just used internally. All library functions and types that are supposed
@@ -37,7 +37,4 @@ SWFFont Ming_loadFont(const char* path, const char* name);
 
 void Ming_cleanupFonts();
 
-float SWFFont_getWideStringWidth(SWFFont font, const unsigned short* string, 
-                                 int len);
-
 #endif /* SWF_FONT_UTIL_H_INCLUDED */
