@@ -19,7 +19,7 @@
 
 /* position.h - wrap matrices in something more usable
  * 
- * $Id: position.h,v 1.9 2003/11/08 18:24:57 whamann Exp $
+ * $Id: position.h,v 1.11 2008/05/31 14:26:08 krechert Exp $
  *
  * Notice: This header file contains declarations of functions and types that
  * are just used internally. All library functions and types that are supposed
@@ -33,31 +33,15 @@
 
 struct SWFPosition_s
 {
-	float x;
-	float y;
-	float xScale;
-	float yScale;
-	float xSkew;
-	float ySkew;
-	float rot;
+	double x;
+	double y;
+	double xScale;
+	double yScale;
+	double xSkew;
+	double ySkew;
+	double rot;
 	SWFMatrix matrix;
 };
 
 
-float SWFPosition_getRotation(SWFPosition position);
-float SWFPosition_getX(SWFPosition position);
-float SWFPosition_getY(SWFPosition position);
-
-void SWFPosition_getXY(SWFPosition position, float* outX, float* outY);
-
-float SWFPosition_getXScale(SWFPosition position);
-float SWFPosition_getYScale(SWFPosition position);
-
-void SWFPosition_getXYScale(SWFPosition position, float* outXScale, float* outYScale);
-
-float SWFPosition_getXSkew(SWFPosition position);
-float SWFPosition_getYSkew(SWFPosition position);
-
-void SWFPosition_getXYSkew(SWFPosition position, float* outXSkew, float* outYSkew);
-
 #endif /* SWF_POSITION_H_INCLUDED */
