@@ -1,3 +1,23 @@
+/****************************************************************************
+ *
+ *  Copyright (C) 2005-2006 "Stuart R. Anderson" <anderson@netsweng.com>
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ ****************************************************************************/
+
 #ifndef _SWFTYPES_H_
 #define _SWFTYPES_H_
 typedef signed char	SI8;	/* Signed  8-bit integer value */
@@ -6,12 +26,15 @@ typedef signed long	SI32;	/* Signed 32-bit integer value */
 typedef unsigned char	UI8;	/* Unsigned  8-bit integer value */
 typedef unsigned short	UI16;	/* Unsigned 16-bit integer value */
 typedef unsigned long	UI32;	/* Unsigned 32-bit integer value */
+typedef unsigned long 	U30;	/* Unsigned 30-bit variable-length encoded integer value */
+typedef unsigned long 	U32;	/* Unsigned 32-bit variable-length encoded integer value */
+typedef signed long	S32;	/* Signed 32-bit variable-length encoded integer value */
 
 typedef unsigned long	FIXED;	/* 32-bit 16.16 fixed-point number */
+typedef unsigned short	FIXED8;	/* 16-bit 8.8 fixed-point number */
 typedef unsigned long	SB32;	/* Signed bit value */
 typedef unsigned long	UB32;	/* Unsigned bit value */
 typedef float		FB32;	/* Signed fixed-point bit value */
-
 typedef char	*STRING;/* String */
 
 typedef UI8 LANGCODE;
@@ -20,13 +43,14 @@ typedef SI16 WORD;
 typedef UI8 BOOLEAN;
 typedef float		FLOAT;	/* 32-bit IEEE single-precision little endian fp value */
 typedef double		DOUBLE;	/* 64-bit IEEE double-precision little endian double value */
+typedef float 		FLOAT16; /* 16-bit IEEE double-precision little endian double value */
 
 /* Types to represent Actions */
 
 struct SWF_ACTIONRECORD {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 };
 
 /* v3 actions */
@@ -34,14 +58,14 @@ struct SWF_ACTIONRECORD {
 struct SWF_ACTIONGOTOFRAME {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	WORD	Frame;
 };
 
 struct SWF_ACTIONGETURL {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	STRING	UrlString;
 	STRING	TargetString;
 };
@@ -73,7 +97,7 @@ struct SWF_ACTIONSTOPSOUNDS {
 struct SWF_ACTIONWAITFORFRAME {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	WORD	Frame;
 	BYTE	SkipCount;
 };
@@ -81,14 +105,14 @@ struct SWF_ACTIONWAITFORFRAME {
 struct SWF_ACTIONSETTARGET {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	STRING	TargetName;
 };
 
 struct SWF_ACTIONGOTOLABEL {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	STRING	FrameLabel;
 };
 
@@ -97,7 +121,7 @@ struct SWF_ACTIONGOTOLABEL {
 struct SWF_ACTIONPUSHPARAM {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI8	Type;
 	union {
 		STRING	String;
@@ -114,7 +138,7 @@ struct SWF_ACTIONPUSHPARAM {
 struct SWF_ACTIONPUSH {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI16	NumParam;
 	struct SWF_ACTIONPUSHPARAM *Params;
 };
@@ -211,14 +235,14 @@ struct SWF_ACTIONMBASCIITOCHAR {
 struct SWF_ACTIONJUMP {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	WORD	BranchOffset;
 };
 
 struct SWF_ACTIONIF {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	WORD	BranchOffset;
 	int	numActions;
 	union SWF_ACTION *Actions;
@@ -227,7 +251,7 @@ struct SWF_ACTIONIF {
 struct SWF_ACTIONCALL {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 };
 
 struct SWF_ACTIONGETVARIABLE {
@@ -241,7 +265,7 @@ struct SWF_ACTIONSETVARIABLE {
 struct SWF_ACTIONGETURL2 {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	union {
 		struct {
 			UI8	SendVarsMethod:2;
@@ -256,7 +280,7 @@ struct SWF_ACTIONGETURL2 {
 struct SWF_ACTIONGOTOFRAME2 {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	union {
 		struct {
 			UI8	Reserved:6;
@@ -299,7 +323,7 @@ struct SWF_ACTIONENDDRAG {
 struct SWF_ACTIONWAITFORFRAME2 {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI8	SkipCount;
 };
 
@@ -329,7 +353,7 @@ struct SWF_ACTIONCALLMETHOD {
 struct SWF_ACTIONCONSTANTPOOL {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI16	Count;
 	STRING	*ConstantPool;
 };
@@ -337,7 +361,7 @@ struct SWF_ACTIONCONSTANTPOOL {
 struct SWF_ACTIONDEFINEFUNCTION {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	STRING	FunctionName;
 	WORD	NumParams;
 	STRING	*Params;
@@ -349,7 +373,7 @@ struct SWF_ACTIONDEFINEFUNCTION {
 struct SWF_ACTIONWITH {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI16	Size;
 	int	numActions;
 	union SWF_ACTION *Actions;
@@ -358,7 +382,7 @@ struct SWF_ACTIONWITH {
 struct SWF_ACTIONSTOREREGISTER {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI8	Register;
 };
 
@@ -392,7 +416,7 @@ struct REGISTERPARAM {
 struct SWF_ACTIONDEFINEFUNCTION2 {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	STRING	FunctionName;
 	UI16	NumParams;
 	UI8	RegisterCount;
@@ -415,7 +439,7 @@ struct SWF_ACTIONDEFINEFUNCTION2 {
 struct SWF_ACTIONTRY {
 	UI8	ActionCode;
 	UI16	Length;
-	UI16	Offset;
+	UI32	Offset;
 	UI8	Reserved:5;
 	UI8	CatchInRegisterFlag:1;
 	UI8	FinallyBlockFlag:1;
@@ -511,6 +535,41 @@ typedef union SWF_ACTION {
 
 /* Types Flash Types */
 
+typedef struct SWF_RGBA {
+	UI8	red;
+	UI8	green;
+	UI8	blue;
+	UI8	alpha;
+} SWF_RGBA;
+
+typedef struct SWF_RECT {
+	UI8	Nbits:5;
+	SB32	Xmin;
+	SB32	Xmax;
+	SB32	Ymin;
+	SB32	Ymax;
+} SWF_RECT;
+
+typedef struct SWF_GRADIENTRECORD {
+	UI8		Ratio;
+	SWF_RGBA	Color;
+} SWF_GRADIENTRECORD;
+
+typedef struct SWF_GRADIENT {
+	UI8	SpreadMode:2;
+	UI8	InterpolationMode:2;
+	UI8	NumGradients:4;
+	SWF_GRADIENTRECORD	GradientRecords[15];
+} SWF_GRADIENT;
+
+typedef struct SWF_FOCALGRADIENT {
+	UI8     SpreadMode:2;
+	UI8     InterpolationMode:2;
+	UI8     NumGradients:4;
+	SWF_GRADIENTRECORD      GradientRecords[15];
+	FIXED8 	FocalPoint;
+} SWF_FOCALGRADIENT;
+
 typedef struct SWF_MATRIX {
 	UI8	HasScale:1;
 	UI8	NScaleBits:5;
@@ -529,30 +588,143 @@ typedef struct SWF_CXFORM {
 	UI8	HasAddTerms:1;
 	UI8	HasMultTerms:1;
 	UI8	Nbits:4;
-	UI32	RedMultTerm;
-	UI32	GreenMultTerm;
-	UI32	BlueMultTerm;
-	UI32	RedAddTerm;
-	UI32	GreenAddTerm;
-	UI32	BlueAddTerm;
+	SI32	RedMultTerm;
+	SI32	GreenMultTerm;
+	SI32	BlueMultTerm;
+	SI32	RedAddTerm;
+	SI32	GreenAddTerm;
+	SI32	BlueAddTerm;
 } SWF_CXFORM;
 
 typedef struct SWF_CXFORMWITHALPHA {
 	UI8	HasAddTerms:1;
 	UI8	HasMultTerms:1;
 	UI8	Nbits:4;
-	UI32	RedMultTerm;
-	UI32	GreenMultTerm;
-	UI32	BlueMultTerm;
-	UI32	AlphaMultTerm;
-	UI32	RedAddTerm;
-	UI32	GreenAddTerm;
-	UI32	BlueAddTerm;
-	UI32	AlphaAddTerm;
+	SI32	RedMultTerm;
+	SI32	GreenMultTerm;
+	SI32	BlueMultTerm;
+	SI32	AlphaMultTerm;
+	SI32	RedAddTerm;
+	SI32	GreenAddTerm;
+	SI32	BlueAddTerm;
+	SI32	AlphaAddTerm;
 } SWF_CXFORMWITHALPHA;
 
+typedef struct SWF_BLURFILTER {
+	FIXED	BlurX;
+	FIXED	BlurY;
+	UI8	Passes:5;
+	UI8	Reserved:3;
+} SWF_BLURFILTER;
+
+typedef struct SWF_BEVELFILTER {
+	SWF_RGBA	ShadowColor;
+	SWF_RGBA	HighlightColor;
+	FIXED		BlurX;
+	FIXED		BlurY;
+	FIXED		Angle;
+	FIXED		Distance;
+	FIXED8		Strength;
+	UI8		InnerShadow:1;
+	UI8		Kockout:1;
+	UI8		CompositeSource:1;
+	UI8		OnTop:1;
+	UI8		Passes:4;
+} SWF_BEVELFILTER;
+
+typedef struct SWF_GRADIENTFILTER
+{
+	UI8		NumColors;
+	SWF_RGBA*	GradientColors;
+	UI8*		GradientRatio;
+	FIXED		BlurX;
+	FIXED		BlurY;
+	FIXED		Angle;
+	FIXED		Distance;
+	FIXED8		Strength;
+	UI8		InnerShadow:1;
+	UI8		Kockout:1;
+	UI8		CompositeSource:1;
+	UI8		OnTop:1;
+	UI8		Passes:4;
+} SWF_GRADIENTFILTER;
+	
+typedef struct SWF_GLOWFILTER {
+	SWF_RGBA 	GlowColor;
+	FIXED		BlurX;
+	FIXED		BlurY;
+	FIXED8		Strength;
+	UI8		InnerGlow:1;
+	UI8		Kockout:1;
+	UI8		CompositeSource:1;
+	UI8		Passes:5;
+} SWF_GLOWFILTER;
+
+typedef struct SWF_CONVOLUTIONFILTER {
+	UI8		MatrixX;
+	UI8		MatrixY;
+	FLOAT		Divisor;
+	FLOAT		Bias;
+	FLOAT*		Matrix;
+	SWF_RGBA	DefaultColor;
+	UI8		Reserved:6;
+	UI8		Clamp:1;
+	UI8		PreserveAlpha:1;
+} SWF_CONVOLUTIONFILTER;
+
+typedef struct SWF_COLORMATRIXFILTER
+{
+	FLOAT Matrix[20];
+} SWF_COLORMATRIXFILTER;
+
+typedef struct SWF_DROPSHADOWFILTER {
+	SWF_RGBA	DropShadowColor;
+	FIXED		BlurX;
+	FIXED		BlurY;
+	FIXED		Angle;
+	FIXED		Distance;
+	FIXED8		Strength;
+	UI8		InnerShadow:1;
+	UI8		Kockout:1;
+	UI8		CompositeSource:1;
+	UI8		Passes:5;
+} SWF_DROPSHADOWFILTER;
+
+enum
+{
+        FILTER_DROPSHADOW,
+        FILTER_BLUR,
+        FILTER_GLOW,
+        FILTER_BEVEL,
+        FILTER_GRADIENTGLOW,
+        FILTER_CONVOLUTION,
+        FILTER_COLORMATRIX,
+        FILTER_GRADIENTBEVEL
+};
+
+typedef struct SWF_FILTER {
+	UI8 	FilterId;
+	union {
+		SWF_DROPSHADOWFILTER dropShadow;
+		SWF_BLURFILTER blur;
+		SWF_GLOWFILTER glow;
+		SWF_BEVELFILTER bevel;
+		SWF_GRADIENTFILTER gradientGlow;
+		SWF_CONVOLUTIONFILTER convolution;
+		SWF_COLORMATRIXFILTER colorMatrix;
+		SWF_GRADIENTFILTER gradientBevel;
+	} filter;
+} SWF_FILTER;
+
+typedef struct SWF_FILTERLIST {
+	UI8 		NumberOfFilters;
+	SWF_FILTER 	*Filter;
+} SWF_FILTERLIST;
+
 typedef struct SWF_BUTTONRECORD {
-	UI8	ButtonReserved:4;
+	UI8	ButtonReserved:2;
+	UI8	ButtonHasBlendMode:1;
+	UI8	ButtonHasFilterList:1;
 	UI8	ButtonStateHitTest:1;
 	UI8	ButtonStateDown:1;
 	UI8	ButtonStateOver:1;
@@ -561,6 +733,8 @@ typedef struct SWF_BUTTONRECORD {
 	UI16	PlaceDepth;
 	SWF_MATRIX	PlaceMatrix;
 	SWF_CXFORMWITHALPHA	ColorTransform;
+	SWF_FILTERLIST	FilterList;	
+	UI8 	BlendMode;
 } SWF_BUTTONRECORD;
 
 typedef struct SWF_BUTTONCONDACTION {
@@ -619,37 +793,12 @@ typedef struct SWF_CLIPACTIONS {
 	UI32	ClipActionEndFlag;
 } SWF_CLIPACTIONS;
 
-
-typedef struct SWF_RGBA {
-	UI8	red;
-	UI8	green;
-	UI8	blue;
-	UI8	alpha;
-} SWF_RGBA;
-
-typedef struct SWF_RECT {
-	UI8	Nbits:5;
-	SB32	Xmin;
-	SB32	Xmax;
-	SB32	Ymin;
-	SB32	Ymax;
-} SWF_RECT;
-
-typedef struct SWF_GRADIENTRECORD {
-	UI8		Ratio;
-	SWF_RGBA	Color;
-} SWF_GRADIENTRECORD;
-
-typedef struct SWF_GRADIENT {
-	UI8	NumGradients;
-	SWF_GRADIENTRECORD	GradientRecords[8];
-} SWF_GRADIENT;
-
 typedef struct SWF_FILLSTYLE {
 	UI8		FillStyleType;
 	SWF_RGBA	Color;
 	SWF_MATRIX	GradientMatrix;
 	SWF_GRADIENT	Gradient;
+	SWF_FOCALGRADIENT FocalGradient;
 	UI16		BitmapId;
 	SWF_MATRIX	BitmapMatrix;
 } SWF_FILLSTYLE;
@@ -665,10 +814,26 @@ typedef struct SWF_LINESTYLE {
 	SWF_RGBA	Color;
 } SWF_LINESTYLE;
 
+typedef struct SWF_LINESTYLE2 {
+	UI16		Width;
+	UI8		StartCapStyle:2;
+	UI8		JoinStyle:2;
+	UI8		HasFillFlag:1;
+	UI8		NoHScaleFlag:1;
+	UI8		NoVScaleFlag:1;
+	UI8		PixelHintingFlag:1;
+	UI8		Reserved:5;
+	UI8		NoClose:1;
+	UI8		EndCapStyle:2;
+	UI16		MiterLimitFactor;
+	SWF_RGBA	Color;
+	SWF_FILLSTYLE	FillType;
+} SWF_LINESTYLE2;
+
 typedef struct SWF_LINESTYLEARRAY {
-	UI8	LineStyleCount;
-	UI16	LineStyleCountExtended;
+	UI16	LineStyleCount;
 	SWF_LINESTYLE	*LineStyles;
+	SWF_LINESTYLE2	*LineStyles2;
 } SWF_LINESTYLEARRAY;
 
 typedef struct SWF_ENDSHAPERECORD {
@@ -724,19 +889,6 @@ typedef union SWF_SHAPERECORD {
 	SWF_CURVEDEDGERECORD	CurvedEdge;
 } SWF_SHAPERECORD;
 
-typedef struct SWF_MORPHLINESTYLE {
-	UI16		StartWidth;
-	UI16		EndWidth;
-	SWF_RGBA	StartColor;
-	SWF_RGBA	EndColor;
-} SWF_MORPHLINESTYLE;
-
-typedef struct SWF_MORPHLINESTYLES {
-	UI8	LineStyleCount;
-	UI16	LineStyleCountExtended;
-	SWF_MORPHLINESTYLE	*LineStyles;
-} SWF_MORPHLINESTYLES;
-
 typedef struct SWF_MORPHGRADIENTRECORD {
 	UI8		StartRatio;
 	SWF_RGBA	StartColor;
@@ -767,6 +919,39 @@ typedef struct SWF_MORPHFILLSTYLES {
 	SWF_MORPHFILLSTYLE	*FillStyles;
 } SWF_MORPHFILLSTYLES;
 
+
+typedef struct SWF_MORPHLINESTYLE {
+	UI16		StartWidth;
+	UI16		EndWidth;
+	SWF_RGBA	StartColor;
+	SWF_RGBA	EndColor;
+} SWF_MORPHLINESTYLE;
+
+typedef struct SWF_MORPHLINESTYLE2 {
+	UI16		StartWidth;
+	UI16		EndWidth;
+	UI16		StartCapStyle:2;
+	UI16		JoinStyle:2;
+	UI16		HasFillFlag:2;
+	UI16		NoHScaleFlag:1;
+	UI16		NoVScaleFlag:1;
+	UI16		PixelHintingFlag:1;
+	UI16		Reserved:5;
+	UI16		NoClose:1;
+	UI16		EndCapStyle:2;
+	UI16		MiterLimitFactor;
+	SWF_RGBA	StartColor;
+	SWF_RGBA	EndColor;
+	SWF_MORPHFILLSTYLE FillType;
+} SWF_MORPHLINESTYLE2;
+
+typedef struct SWF_MORPHLINESTYLES {
+	UI8	LineStyleCount;
+	UI16	LineStyleCountExtended;
+	SWF_MORPHLINESTYLE	*LineStyles;
+	SWF_MORPHLINESTYLE2	*LineStyles2;
+} SWF_MORPHLINESTYLES;
+
 typedef struct SWF_SHAPE {
 	UI8	NumFillBits:4;
 	UI8	NumLineBits:4;
@@ -804,6 +989,28 @@ typedef struct SWF_TEXTRECORD {
 	SWF_GLYPHENTRY *GlyphEntries;
 } SWF_TEXTRECORD;
 
+typedef struct SWF_SOUNDENVELOPE {
+	UI32	Pos44;
+	UI16	LeftLevel;
+	UI16	RightLevel;
+} SWF_SOUNDENVELOPE;
+
+typedef struct SWF_SOUNDINFO {
+	UI8	Reserved:2;
+	UI8	SyncStop:1;
+	UI8	SyncNoMultiple:1;
+	UI8	HasEnvelope:1;
+	UI8	HasLoops:1;
+	UI8	HasOutPoint:1;
+	UI8	HasInPoint:1;
+	UI32	InPoint;
+	UI32	OutPoint;
+	UI16	LoopCount;
+	UI8	EnvPoints;
+	SWF_SOUNDENVELOPE *EnvelopeRecords;
+} SWF_SOUNDINFO;
+
+
 
 /* Types to represent Blocks */
 
@@ -835,6 +1042,7 @@ struct SWF_DEFINEBITSJPEG3
   UI8 *BitmapAlphaData;
 };
 
+// XXX: obsolete
 struct SWF_DEFINEBITSPTR
 {
   int chid;
@@ -842,7 +1050,13 @@ struct SWF_DEFINEBITSPTR
 
 struct SWF_DEFINEBUTTON
 {
-  int chid;
+  UI16 ButtonId;
+  SWF_BUTTONRECORD *Characters;
+  int numCharacters;
+  UI8 CharacterEndFlag;
+  int numActions;
+  SWF_ACTION *Actions;
+  UI8 ActionEndFlag;
 };
 
 struct SWF_DEFINEBUTTON2
@@ -860,14 +1074,25 @@ struct SWF_DEFINEBUTTON2
 
 struct SWF_DEFINEBUTTONCXFORM
 {
-  int chid;
+  UI16 ButtonId;
+  SWF_CXFORM ButtonColorTransform;
 };
 
 struct SWF_DEFINEBUTTONSOUND
 {
-  int chid;
+  UI16 CharacterID;
+  UI16 ButtonSoundChar0;
+  SWF_SOUNDINFO ButtonSoundInfo0;
+  UI16 ButtonSoundChar1;
+  SWF_SOUNDINFO ButtonSoundInfo1;
+  UI16 ButtonSoundChar2;
+  SWF_SOUNDINFO ButtonSoundInfo2;
+  UI16 ButtonSoundChar3;
+  SWF_SOUNDINFO ButtonSoundInfo3;
 };
 
+
+// XXX: obsolete ?
 struct SWF_DEFINECOMMANDOBJ
 {
   int chid;
@@ -885,15 +1110,16 @@ struct SWF_DEFINEEDITTEXT
   UI8	HasTextColor:1;
   UI8	HasMaxLength:1;
   UI8	HasFont:1;
-  UI8	Reserved:1;
+  UI8	HasFontClass:1;
   UI8	AutoSize:1;
   UI8	HasLayout:1;
   UI8	NoSelect:1;
   UI8	Border:1;
-  UI8	Reserved2:1;
+  UI8	WasStatic:1;
   UI8	HTML:1;
   UI8	UseOutlines:1;
   UI16	FontID;
+  STRING FontClass;
   UI16	FontHeight;
   SWF_RGBA	TextColor;
   UI16	MaxLength;
@@ -909,15 +1135,16 @@ struct SWF_DEFINEEDITTEXT
 struct SWF_DEFINEFONT
 {
   UI16 FontID;
+  int NumGlyphs;
   UI16 *OffsetTable;
   SWF_SHAPE *GlyphShapeTable;
 };
 
 struct SWF_KERNINGRECORD
 {
-  UI8 FontKerningCode1;
-  UI8 FontKerningCode2;
-  SI16 FontKerningADjustment;
+  UI16 FontKerningCode1;
+  UI16 FontKerningCode2;
+  SI16 FontKerningAdjustment;
 };
 
 struct SWF_DEFINEFONT2
@@ -954,6 +1181,40 @@ struct SWF_DEFINEFONT2
   struct SWF_KERNINGRECORD *FontKerningTable;
 };
 
+struct SWF_DEFINEFONT3
+{
+  UI16 FontID;
+  UI8 FontFlagsHasLayout:1;
+  UI8 FontFlagsShiftJis:1;
+  UI8 FontFlagsSmallText:1;
+  UI8 FontFlagsFlagANSI:1;
+  UI8 FontFlagsWideOffsets:1;
+  UI8 FontFlagsWideCodes:1;
+  UI8 FontFlagsFlagsItalics:1;
+  UI8 FontFlagsFlagsBold:1;
+  LANGCODE LanguageCode;
+  UI8 FontNameLen;
+  STRING FontName;
+  UI16 NumGlyphs;
+  union {
+    UI16 *UI16;
+    UI32 *UI32;
+  } OffsetTable;
+  union {
+    UI16 UI16;
+    UI32 UI32;
+  } CodeTableOffset;
+  SWF_SHAPE *GlyphShapeTable;
+  UI16 *CodeTable;
+  SI16 FontAscent;
+  SI16 FontDecent;
+  SI16 FontLeading;
+  SI16 *FontAdvanceTable;
+  SWF_RECT *FontBoundsTable;
+  UI16 KerningCount;
+  struct SWF_KERNINGRECORD *FontKerningTable;
+};
+
 struct SWF_DEFINEFONTINFO
 {
   UI16 FontID;
@@ -970,6 +1231,65 @@ struct SWF_DEFINEFONTINFO
   UI16 *CodeTable;
 };
 
+struct SWF_DEFINEFONTINFO2
+{
+  UI16 FontID;
+  UI8 FontNameLen;
+  STRING FontName;
+  UI8 FontFlagsReserved:2;
+  UI8 FontFlagsSmallText:1;
+  UI8 FontFlagsShiftJIS:1;
+  UI8 FontFlagsANSI:1;
+  UI8 FontFlagsItalic:1;
+  UI8 FontFlagsBold:1;
+  UI8 FontFlagsWideCodes:1;
+  LANGCODE LanguageCode;
+  int nGlyph;
+  UI16 *CodeTable;
+};
+
+struct SWF_ZONEDATA
+{
+  FLOAT16 AlignmentCoordinate;
+  FLOAT16 Range;
+};
+
+struct SWF_ZONERECORD
+{
+  UI8 NumZoneData;
+  struct SWF_ZONEDATA *ZoneData;
+  UI8 ZoneMaskX:1;
+  UI8 ZoneMaskY:1;
+  UI8 Reserved:6;
+};
+
+struct SWF_DEFINEFONTALIGNZONES
+{
+  UI16 FontID;
+  UI8 CSMTableHint:2;
+  UI8 Reserved:6;
+  int GlyphCount;
+  struct SWF_ZONERECORD *ZoneTable;
+};
+
+struct SWF_CSMTEXTSETTINGS
+{
+  UI16 TextID;
+  UI8 UseFlashType:2;
+  UI8 GridFit:3;
+  UI8 Reserved:3;
+  FB32 Thickness;
+  FB32 Sharpness;
+  UI8 Reserved2;
+};
+
+struct SWF_DEFINEFONTNAME
+{
+  UI16 FontId;
+  STRING FontName;
+  STRING FontCopyright;
+};
+
 struct SWF_DEFINELOSSLESS
 {
   UI16 CharacterID;
@@ -1002,6 +1322,24 @@ struct SWF_DEFINEMORPHSHAPE
   SWF_SHAPE EndEdges;
 };
 
+struct SWF_DEFINEMORPHSHAPE2
+{
+  UI16 CharacterID;
+  SWF_RECT StartBounds;
+  SWF_RECT EndBounds;
+  SWF_RECT StartEdgeBounds;
+  SWF_RECT EndEdgeBounds;
+  UI8 Reserved:6;
+  UI8 UsesNonScalingStrokes:1;
+  UI8 UsesScalingStrokes:1;
+  UI32 Offset;
+  SWF_MORPHFILLSTYLES MorphFillStyles;
+  SWF_MORPHLINESTYLES MorphLineStyles;
+  SWF_SHAPE StartEdges;
+  SWF_SHAPE EndEdges;
+};
+
+
 struct SWF_DEFINESHAPE
 {
   UI16 ShapeID;
@@ -1023,15 +1361,15 @@ struct SWF_DEFINESHAPE3
   SWF_SHAPEWITHSTYLE Shapes;
 };
 
-struct SWF_DEFINESOUND
+struct SWF_DEFINESHAPE4
 {
-  UI16 SoundId;
-  UI8  SoundFormat:4;
-  UI8  SoundRate:2;
-  UI8  SoundSize:1;
-  UI8  SoundType:1;
-  UI32  SoundSampleCount;
-  UI8  *SoundData;
+  UI16 ShapeID;
+  SWF_RECT ShapeBounds;
+  SWF_RECT EdgeBounds;
+  UI8 Reserved:6;
+  UI8 UsesNonScalingStrokes:1;
+  UI8 UsesScalingStrokes:1;
+  SWF_SHAPEWITHSTYLE Shapes;
 };
 
 struct SWF_DEFINESPRITE
@@ -1040,7 +1378,7 @@ struct SWF_DEFINESPRITE
   UI16 FrameCount;
   UI16 BlockCount;
   UI16 *tagTypes;
-  union SWF_Parserstruct **Tags;
+  struct SWF_Parserstruct_s **Tags;
 };
 
 struct SWF_DEFINETEXT
@@ -1065,11 +1403,13 @@ struct SWF_DEFINETEXT2
   SWF_TEXTRECORD *TextRecords;
 };
 
+// XXX obsolete ?
 struct SWF_DEFINETEXTFORMAT
 {
   int chid;
 };
 
+// XXX obsolete ?
 struct SWF_DEFINEVIDEO
 {
   int chid;
@@ -1077,7 +1417,14 @@ struct SWF_DEFINEVIDEO
 
 struct SWF_DEFINEVIDEOSTREAM
 {
-  int chid;
+  UI16 CharacterID;
+  UI16 NumFrames;
+  UI16 Width;
+  UI16 Height;
+  UI8 Reserved:5;
+  UI8 VideoFlagsDeblocking:2;
+  UI8 VideoFlagsSmoothing:1;
+  UI8 CodecID;
 };
 
 struct SWF_DOACTION
@@ -1088,13 +1435,17 @@ struct SWF_DOACTION
 
 struct SWF_ENABLEDEBUGGER
 {
-  int chid;
+  STRING Password;
+};
+
+struct SWF_ENABLEDEBUGGER2
+{
+  UI16 Reserved;
+  STRING Password;
 };
 
 struct SWF_END
-{
-  int chid;
-};
+{ /* empty tag */ };
 
 struct SWF_EXPORTASSETS
 {
@@ -1103,6 +1454,7 @@ struct SWF_EXPORTASSETS
   STRING *Names;
 };
 
+// XXX obsolete ?
 struct SWF_FONTREF
 {
   int chid;
@@ -1111,23 +1463,28 @@ struct SWF_FONTREF
 struct SWF_FRAMELABEL
 {
   STRING	Name;
+  UI8		IsAnchor;
 };
 
+// XXX obsolete ?
 struct SWF_FRAMETAG
 {
   int chid;
 };
 
+// XXX obsolete ?
 struct SWF_FREEALL
 {
   int chid;
 };
 
+// XXX obsolete ?
 struct SWF_FREECHARACTER
 {
   int chid;
 };
 
+// XXX obsolete ?
 struct SWF_GENCOMMAND
 {
   int chid;
@@ -1141,6 +1498,17 @@ struct SWF_IMPORTASSETS
   STRING *Names;
 };
 
+struct SWF_IMPORTASSETS2
+{
+  STRING URL;
+  UI8 Reserved;
+  UI8 Reserved2;
+  UI16	Count;
+  UI16 *Tags;
+  STRING *Names;
+};
+
+
 struct SWF_JPEGTABLES
 {
   int JPEGDataSize;
@@ -1149,9 +1517,11 @@ struct SWF_JPEGTABLES
 
 struct SWF_NAMECHARACTER
 {
-  int chid;
+  UI16 Id;
+  STRING Name;
 };
 
+// XXX: obsolete?
 struct SWF_PATHSAREPOSTSCRIPT
 {
   int chid;
@@ -1159,7 +1529,10 @@ struct SWF_PATHSAREPOSTSCRIPT
 
 struct SWF_PLACEOBJECT
 {
-  int chid;
+  UI16 CharacterId;
+  UI16 Depth;
+  SWF_MATRIX Matrix;
+  SWF_CXFORMWITHALPHA ColorTransform;
 };
 
 struct SWF_PLACEOBJECT2
@@ -1182,6 +1555,35 @@ struct SWF_PLACEOBJECT2
   SWF_CLIPACTIONS ClipActions;
 };
 
+struct SWF_PLACEOBJECT3
+{
+  UI8 PlaceFlagHasClipActions:1;
+  UI8 PlaceFlagHasClipDepth:1;
+  UI8 PlaceFlagHasName:1;
+  UI8 PlaceFlagHasRatio:1;
+  UI8 PlaceFlagHasColorTransform:1;
+  UI8 PlaceFlagHasMatrix:1;
+  UI8 PlaceFlagHasCharacter:1;
+  UI8 PlaceFlagMove:1;
+  UI8 Reserved:3;
+  UI8 PlaceFlagHasImage:1;
+  UI8 PlaceFlagHasClassName:1;
+  UI8 PlaceFlagHasCacheAsBitmap:1;
+  UI8 PlaceFlagHasBlendMode:1;
+  UI8 PlaceFlagHasFilterList:1;
+  UI16 Depth;
+  STRING ClassName;
+  UI16 CharacterId;
+  SWF_MATRIX Matrix;
+  SWF_CXFORMWITHALPHA ColorTransform;
+  UI16 Ratio;
+  STRING Name;
+  UI16 ClipDepth;
+  SWF_FILTERLIST SurfaceFilterList;
+  UI8 BlendMode;
+  SWF_CLIPACTIONS ClipActions;
+};
+
 struct SWF_PREBUILT
 {
   int chid;
@@ -1194,7 +1596,7 @@ struct SWF_PREBUILTCLIP
 
 struct SWF_PROTECT
 {
-  int chid;
+  STRING Password;
 };
 
 struct SWF_REMOVEOBJECT
@@ -1209,8 +1611,15 @@ struct SWF_REMOVEOBJECT2
 };
 
 struct SWF_SERIALNUMBER
-{
-  int chid;
+{ 
+  UI32 Id;
+  UI32 Edition;
+  UI8 Major;
+  UI8 Minor;
+  UI32 BuildL;
+  UI32 BuildH;
+  UI32 TimestampL;
+  UI32 TimestampH;
 };
 
 struct SWF_SETBACKGROUNDCOLOR
@@ -1223,9 +1632,19 @@ struct SWF_SHOWFRAME
   int chid;
 };
 
+struct MP3STREAMSOUNDDATA
+{
+  SI16 SampleCount;
+  SI16 SeekSamples;
+  UI8 *frames;
+};
+
 struct SWF_SOUNDSTREAMBLOCK
 {
-  int chid;
+  union {
+    struct MP3STREAMSOUNDDATA mp3;
+    UI8 *data;
+  } StreamData;
 };
 
 struct SWF_SOUNDSTREAMHEAD
@@ -1256,11 +1675,34 @@ struct SWF_SOUNDSTREAMHEAD2
   UI16	LatencySeek;
 };
 
-struct SWF_STARTSOUND
+struct SWF_DEFINESOUND
 {
-  int chid;
+  UI16 SoundId;
+  UI8  SoundFormat:4;
+  UI8  SoundRate:2;
+  UI8  SoundSize:1;
+  UI8  SoundType:1;
+  UI32  SoundSampleCount;
+  union {
+    struct MP3STREAMSOUNDDATA mp3;
+    UI8 *data;
+  } SoundData;
 };
 
+struct SWF_STARTSOUND
+{
+  UI16 SoundId;
+  SWF_SOUNDINFO	SoundInfo;
+};
+
+struct SWF_STARTSOUND2
+{
+  STRING SoundClassName;
+  SWF_SOUNDINFO	SoundInfo;
+};
+
+
+// XXX: obsolete?
 struct SWF_SYNCFRAME
 {
   int chid;
@@ -1275,11 +1717,341 @@ struct SWF_INITACTION
 
 struct SWF_VIDEOFRAME
 {
-  int chid;
+  UI16 StreamID;
+  UI16 FrameNum;
+  UI8 *VideoData;
+};
+
+struct SWF_REFLEX
+{
+  UI8 rfx[3];
+};
+
+struct SWF_FILEATTRIBUTES
+{
+	UI8	Reserved:3;
+	UI8	HasMetadata:1;
+	UI8	ActionScript3:1;
+	UI8	Reserved2:2;
+	UI8	UseNetwork:1;
+	UI16	Reserved3;
+	UI8	Reserved4;
+};
+
+struct SWF_METADATA
+{
+	STRING Metadata;
+};
+
+struct SWF_SCRIPTLIMITS
+{
+	UI16 MaxRecursionDepth;
+	UI16 ScriptTimeoutSeconds;
+};
+
+struct SWF_DEFINESCALINGGRID
+{
+	UI16 CharacterId;
+	SWF_RECT Splitter;
+};
+
+struct SWF_SETTABINDEX
+{
+	UI16 Depth;
+	UI16 TabIndex;
+};
+
+struct ABC_OPTION_DETAIL
+{
+	U30 Val;
+	UI8 Kind;
+};
+
+struct ABC_OPTION_INFO
+{
+	U30 OptionCount;
+	struct ABC_OPTION_DETAIL *Option;	
+};
+
+struct ABC_PARAM_INFO
+{
+	U30 *ParamNames;
+};
+
+struct ABC_METHOD_INFO
+{
+	U30 ParamCount;
+	U30 ReturnType;
+	U30 *ParamType;
+	U30 Name;
+	UI8 Flags;
+	struct ABC_OPTION_INFO Options;
+	struct ABC_PARAM_INFO ParamNames;
+};	
+
+struct ABC_STRING_INFO
+{
+	U30 Size;
+	UI8 *UTF8String;
+};
+
+struct ABC_NS_INFO
+{
+	UI8 Kind;
+	U30 Name;
+};
+
+struct ABC_NS_SET_INFO
+{
+	U30 Count;
+	U30 *NS;
+};
+
+struct ABC_QNAME
+{
+	U30 NS;
+	U30 Name;
+};
+
+struct ABC_RTQNAME
+{
+	U30 Name;
+};
+
+struct ABC_RTQNAME_L
+{
+	// empty
+};
+
+struct ABC_MULTINAME
+{
+	U30 Name;
+	U30 NSSet;
+};
+
+struct ABC_MULTINAME_L
+{
+	U30 NSSet;
+};
+
+struct ABC_MULTINAME_INFO
+{
+	UI8 Kind;
+	union {
+		struct ABC_QNAME QName;
+		struct ABC_QNAME QNameA;
+		struct ABC_RTQNAME RTQName;
+		struct ABC_RTQNAME RTQNameA;
+		struct ABC_RTQNAME_L RTQNameL;
+		struct ABC_RTQNAME_L RTQNameLA;
+		struct ABC_MULTINAME Multiname;
+		struct ABC_MULTINAME MultinameA;
+		struct ABC_MULTINAME_L MultinameL;
+		struct ABC_MULTINAME_L MultinameLA;
+	} Data;
+
+};
+
+struct ABC_CONSTANT_POOL
+{
+	U30 IntCount;
+	S32 *Integers;
+	U30 UIntCount;
+	U32 *UIntegers;
+	U30 DoubleCount;
+	DOUBLE *Doubles;
+	U30 StringCount;
+	struct ABC_STRING_INFO *Strings;
+	U30 NamespaceCount;
+	struct ABC_NS_INFO *Namespaces;
+	U30 NamespaceSetCount;
+	struct ABC_NS_SET_INFO *NsSets;
+	U30 MultinameCount;
+	struct ABC_MULTINAME_INFO *Multinames;
+};
+
+struct ABC_ITEM_INFO
+{
+	U30 Key;
+	U30 Value;
+};
+
+struct ABC_METADATA_INFO
+{
+	U30 Name;
+	U30 ItemCount;
+	struct ABC_ITEM_INFO *Items;
 };
 
 
-typedef union SWF_Parserstruct
+struct ABC_TRAIT_SLOT
+{
+	U30 SlotId;
+	U30 TypeName;
+	U30 VIndex;
+	UI8 VKind;
+
+};
+
+struct ABC_TRAIT_CLASS
+{
+	U30 SlotId;
+	U30 ClassIndex;
+};
+
+struct ABC_TRAIT_FUNCTION
+{
+	U30 SlotId;
+	U30 Function;
+};
+
+struct ABC_TRAIT_METHOD
+{
+	U30 DispId;
+	U30 Method;
+};
+
+struct ABC_TRAITS_INFO
+{
+	U30 Name;
+	UI8 Kind;
+	UI8 Attr;
+	union {
+		struct ABC_TRAIT_SLOT Slot;
+		struct ABC_TRAIT_CLASS Class;
+		struct ABC_TRAIT_FUNCTION Function;
+		struct ABC_TRAIT_METHOD Method;
+	} Data;
+	U30 MetadataCount;
+	U30 *Metadata;
+};
+
+struct ABC_INSTANCE_INFO
+{
+	U30 Name;
+	U30 SuperName;
+	UI8 Flags;
+	U30 ProtectedNs;
+	U30 InterfaceCount;
+	U30 *Interfaces;
+	U30 IInit;
+	U30 TraitCount;
+	struct ABC_TRAITS_INFO *Traits;
+};	
+
+struct ABC_CLASS_INFO
+{
+	U30 CInit;
+	U30 TraitCount;
+	struct ABC_TRAITS_INFO *Traits;
+};
+
+struct ABC_SCRIPT_INFO
+{
+	U30 Init;
+	U30 TraitCount;
+	struct ABC_TRAITS_INFO *Traits;
+};
+
+struct ABC_EXCEPTION_INFO
+{
+	U30 From;
+	U30 To;
+	U30 Target;
+	U30 ExcType;
+	U30 VarName;
+};
+
+struct ABC_METHOD_BODY_INFO
+{
+	U30 Method;
+	U30 MaxStack;
+	U30 LocalCount;
+	U30 InitScopeDepth;
+	U30 MaxScopeDepth;
+	U30 CodeLength;
+	UI8 *Code;
+	U30 ExceptionCount;
+	struct ABC_EXCEPTION_INFO *Exceptions;
+	U30 TraitCount;
+	struct ABC_TRAITS_INFO *Traits;
+};
+
+struct ABC_FILE
+{
+	UI16 Minor;
+	UI16 Major;
+	struct ABC_CONSTANT_POOL ConstantPool;
+	U30 MethodCount;
+	struct ABC_METHOD_INFO *Methods;
+	U30 MetadataCount;
+	struct ABC_METADATA_INFO *Metadata;
+	U30 ClassCount;
+	struct ABC_INSTANCE_INFO *Instances;
+	struct ABC_CLASS_INFO *Classes;
+	U30 ScriptCount;
+	struct ABC_SCRIPT_INFO *Scripts;
+	U30 MethodBodyCount;
+	struct ABC_METHOD_BODY_INFO *MethodBodies;	
+};
+
+struct SWF_DOABC
+{
+	UI32 Flags;
+	STRING Name;
+	struct ABC_FILE AbcFile;
+};
+
+struct AS_SYMBOL
+{
+	UI16 SymbolId;
+	STRING SymbolName;
+};
+
+struct SWF_SYMBOLCLASS
+{
+	UI16 SymbolCount;
+	struct AS_SYMBOL *SymbolList;
+};
+
+struct SWF_DEFINEBINARYDATA
+{	
+	UI32 Reserved;
+	int DataLength;
+	UI8 *Data;
+};
+
+struct SCENE_DATA
+{
+	U32 Offset;
+	STRING Name;
+};
+
+struct FRAME_DATA
+{
+	U32 FrameNum;
+	STRING FrameLabel;
+};
+
+struct SWF_DEFINESCENEANDFRAMEDATA
+{
+	U32 SceneCount;
+	struct SCENE_DATA *Scenes;
+	U32 FrameLabelCount;
+	struct FRAME_DATA *Frames;
+};
+
+struct SWF_DEBUGID
+{
+	UI8* UUID;
+};
+
+struct SWF_UNKNOWNBLOCK
+{
+	UI8 *Data;
+};
+
+typedef union _inline_SWF_Parserstruct
 {
   struct SWF_CHARACTERSET SWF_CHARACTERSET;
   struct SWF_DEFINEBITS SWF_DEFINEBITS;
@@ -1294,13 +2066,20 @@ typedef union SWF_Parserstruct
   struct SWF_DEFINEEDITTEXT SWF_DEFINEEDITTEXT;
   struct SWF_DEFINEFONT SWF_DEFINEFONT;
   struct SWF_DEFINEFONT2 SWF_DEFINEFONT2;
+  struct SWF_DEFINEFONT3 SWF_DEFINEFONT3;
   struct SWF_DEFINEFONTINFO SWF_DEFINEFONTINFO;
+  struct SWF_DEFINEFONTINFO2 SWF_DEFINEFONTINFO2;
+  struct SWF_DEFINEFONTALIGNZONES SWF_DEFINEFONTALIGNZONES;
+  struct SWF_CSMTEXTSETTINGS SWF_CSMTEXTSETTINGS;
+  struct SWF_DEFINEFONTNAME SWF_DEFINEFONTNAME;
   struct SWF_DEFINELOSSLESS SWF_DEFINELOSSLESS;
   struct SWF_DEFINELOSSLESS2 SWF_DEFINELOSSLESS2;
   struct SWF_DEFINEMORPHSHAPE SWF_DEFINEMORPHSHAPE;
+  struct SWF_DEFINEMORPHSHAPE2 SWF_DEFINEMORPHSHAPE2;
   struct SWF_DEFINESHAPE SWF_DEFINESHAPE;
   struct SWF_DEFINESHAPE2 SWF_DEFINESHAPE2;
   struct SWF_DEFINESHAPE3 SWF_DEFINESHAPE3;
+  struct SWF_DEFINESHAPE4 SWF_DEFINESHAPE4;
   struct SWF_DEFINESOUND SWF_DEFINESOUND;
   struct SWF_DEFINESPRITE SWF_DEFINESPRITE;
   struct SWF_DEFINETEXT SWF_DEFINETEXT;
@@ -1310,6 +2089,7 @@ typedef union SWF_Parserstruct
   struct SWF_DEFINEVIDEOSTREAM SWF_DEFINEVIDEOSTREAM;
   struct SWF_DOACTION SWF_DOACTION;
   struct SWF_ENABLEDEBUGGER SWF_ENABLEDEBUGGER;
+  struct SWF_ENABLEDEBUGGER2 SWF_ENABLEDEBUGGER2;
   struct SWF_END SWF_END;
   struct SWF_EXPORTASSETS SWF_EXPORTASSETS;
   struct SWF_FONTREF SWF_FONTREF;
@@ -1319,11 +2099,13 @@ typedef union SWF_Parserstruct
   struct SWF_FREECHARACTER SWF_FREECHARACTER;
   struct SWF_GENCOMMAND SWF_GENCOMMAND;
   struct SWF_IMPORTASSETS SWF_IMPORTASSETS;
+  struct SWF_IMPORTASSETS2 SWF_IMPORTASSETS2;
   struct SWF_JPEGTABLES SWF_JPEGTABLES;
   struct SWF_NAMECHARACTER SWF_NAMECHARACTER;
   struct SWF_PATHSAREPOSTSCRIPT SWF_PATHSAREPOSTSCRIPT;
   struct SWF_PLACEOBJECT SWF_PLACEOBJECT;
   struct SWF_PLACEOBJECT2 SWF_PLACEOBJECT2;
+  struct SWF_PLACEOBJECT3 SWF_PLACEOBJECT3;
   struct SWF_PREBUILT SWF_PREBUILT;
   struct SWF_PREBUILTCLIP SWF_PREBUILTCLIP;
   struct SWF_PROTECT SWF_PROTECT;
@@ -1336,8 +2118,30 @@ typedef union SWF_Parserstruct
   struct SWF_SOUNDSTREAMHEAD SWF_SOUNDSTREAMHEAD;
   struct SWF_SOUNDSTREAMHEAD2 SWF_SOUNDSTREAMHEAD2;
   struct SWF_STARTSOUND SWF_STARTSOUND;
+  struct SWF_STARTSOUND2 SWF_STARTSOUND2;
   struct SWF_SYNCFRAME SWF_SYNCFRAME;
   struct SWF_INITACTION SWF_INITACTION;
   struct SWF_VIDEOFRAME SWF_VIDEOFRAME;
-} SWF_Parserstruct;
+  struct SWF_REFLEX SWF_REFLEX;
+  struct SWF_FILEATTRIBUTES SWF_FILEATTRIBUTES;
+  struct SWF_METADATA SWF_METADATA;
+  struct SWF_SCRIPTLIMITS SWF_SCRIPTLIMITS;
+  struct SWF_DEFINESCALINGGRID SWF_DEFINESCALINGGRID;
+  struct SWF_SETTABINDEX SWF_SETTABINDEX;
+  struct SWF_DOABC SWF_DOABC;
+  struct SWF_SYMBOLCLASS SWF_SYMBOLCLASS;
+  struct SWF_DEFINEBINARYDATA SWF_DEFINEBINARYDATA;
+  struct SWF_DEFINESCENEANDFRAMEDATA SWF_DEFINESCENEANDFRAMEDATA;
+  struct SWF_DEBUGID SWF_DEBUGID;
+  struct SWF_UNKNOWNBLOCK SWF_UNKNOWNBLOCK;
+} _inline_SWF_Parserstruct;
+
+struct SWF_Parserstruct_s
+{
+	_inline_SWF_Parserstruct block;
+	int offset;
+	int length;
+};
+typedef struct SWF_Parserstruct_s SWF_Parserstruct; 
+
 #endif /* _SWFTYPES_H_ */
